"""Minipipe is a machine learning mini-batch pipeline tool for out-of-memory workflows. """

from .base import Sentinel
from .base import Logger
from .base import Stream
from .base import Pipe
from .pipelines import PipeSystem
from .pipelines import PipeLine
from .pipes import Source
from .pipes import Sink
from .pipes import Transform
from .pipes import Regulator



"""Base classes"""

from multiprocessing import Queue, Event
from queue import Empty, Full
import logging

class Sentinel(object):

    """
        This class is used to indicate the end of a stream. When a instance of Sentinel is
        passed to a Pipe it will shut itself down.
    """
    def __repr__(self):
        return 'sentinel'

class Logger(object):

    """
        Logger class used by Pipe. There are five levels of logs: INFO, DEBUG, WARNING, ERROR and CRITICAL.
        By default logger is set to INFO.

        :param lvl: log level, one of: info, debug, warning, error or critical
        :return: None
    """

    def __init__(self, lvl='INFO'):
        self.log_lvl_map = {'INFO':logging.INFO,
                            'DEBUG':logging.DEBUG,
                            'WARNING':logging.WARNING,
                            'ERROR':logging.ERROR,
                            'CRITICAL':logging.CRITICAL}

        assert(lvl in self.log_lvl_map), "log_lvl must be one of: {}".format(self.log_lvl_map.keys())
        self.lvl = self.log_lvl_map[lvl]
        self.logger = logging.getLogger("logger")
        self.logger.propagate = False

        # jupyter already has a logger initialized
        # this avoids initializing it multiple times
        if hasattr(self.logger, 'handler_set'):
            while self.logger.handlers:
                self.logger.handlers.pop()
        ch = logging.StreamHandler()
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(pname)s - %(message)s')
        ch.setFormatter(formatter)
        ch.setLevel(lvl)
        self.logger.addHandler(ch)
        self.logger.setLevel(lvl)
        self.logger.handler_set = True

    def log(self, msg, name, lvl='info'):

        """
            Logs messages.

            :param msg: message to log
            :param name: name of object invoking logs
            :param lvl: log level, one of: info, debug, warning, error or critical
            :return: None
        """

        extra = {'pname':name}
        if lvl == 'critical':
            self.logger.critical(msg, extra=extra)
        elif lvl == 'debug':
            self.logger.debug(msg, extra=extra)
        elif lvl == 'warning':
            self.logger.warning(msg, extra=extra)
        elif lvl == 'error':
            self.logger.error(msg, extra=extra)
        else:
            self.logger.info(msg, extra=extra)


class Stream(object):

    """
        Based off of a multiprocessing queue, Stream handles moving data between Pipe segments.
    """

    def __init__(self, name = 'stream', buffer_size=3, timeout=None, monitor=False):
        self.q = Queue(buffer_size)
        self.timeout = timeout
        self.buffer_size = buffer_size
        self.name = name
        self.monitor = monitor
        self.pipes_out = [] # Reference to Pipes that put onto this stream
        self.pipes_in = []  # Reference to Pipes that get from this stream
        self.logger = None

    def _id(self):
        return str(id(self))

    def __hash__(self):
        return int(self._id())

    def __eq__(self, other):
        return self.__hash__() == hash(other)

    def add_pipe_in(self, pipe_in):
        self.pipes_in.append(pipe_in)

    def add_pipe_out(self, pipe_out):
        self.pipes_out.append(pipe_out)

    def empty(self):
        return self.q.empty()

    def full(self):
        return self.q.full()

    def capacity(self):
        return 100.0*self.q.qsize()/self.buffer_size

    def set_logger(self, logger):
        self.logger = logger

    def flush(self):
        try:
            while True:
                self.q.get_nowait()
        except Empty:
            pass

    def get(self, timeout=1):

        # While pipes out are running and not empty try to get data
        # timeout value is necessary else get may run indefinitely after pipes out shutdown

        x = None
        while (any([p._continue() for p in self.pipes_out]) or not self.q.empty()):
            try:
                x = self.q.get(timeout=timeout or self.timeout)
                if self.monitor:
                    self.logger.log('capacity:{:0.2f}%'.format(self.capacity()),
                                    self.name+':get')
                break
            except Empty:
                continue
        return x

    def put(self, x, timeout=1):

        # While pipes in are running try to put data
        # no need to check if full since Full exception is caught
        # timeout value is necessary else put may run indefinitely after pipes in shutdown

        while any([p._continue() for p in self.pipes_in]):
            try:
                if x is None:
                    break
                self.q.put(x, timeout=timeout or self.timeout)
                if self.monitor:
                    self.logger.log('capacity:{:0.2f}%'.format(self.capacity()),
                                    self.name+':put')
                break
            except Full:
                continue

class Pipe(object):

    """
        Base class for all pipe segments. Pipes use two sets of Streams: upstreams and downstreams.
        Generally Pipes except data from upstreams and pass downstream after a transformation.
        All pipe segments run on their own thread or process, which allows them to run in
        parallel with other segments.

        Number of upstreams should be equal to number of functor args. Likewise, number of downstreams
        should be equal to number of functor outputs.

        When Pipe produces a None it will not be passed downstream. In this case nothing will be placed
        on the downstreams. This allows the user to create 'switches' based on internal logic in the functor.


        Base initializer
        -----------------

        :param functor: Python function, class, generator or corountine
        :param name: String associated with pipe segment
        :param upstreams: List of Streams that are inputs to functor
        :param downstreams: List of Streams that are outputs of functor
        :param ignore_exceptions: List of exceptions to ignore while pipeline is running
        :param init_kwargs: Kwargs to initiate class object on process (no used when func_type = 'function')
        :param stateful: Set to True when using a class functor. Class functors must implement a 'run' method

    """

    def __init__(self, functor_obj, name, upstreams=None, downstreams=None,
                 ignore_exceptions=None, init_kwargs=None):

        # Public methods
        self.functor_obj = functor_obj
        self.name = name
        self.init_kwargs = init_kwargs or {}
        self.upstreams = upstreams or []
        self.downstreams = downstreams or []
        self.ignore_exceptions = ignore_exceptions or []

        # Private methods
        self._n_desc = 0
        self._n_ances = 0
        self._n_rcvd_term_sigs = 0
        self._n_outputs = len(self.downstreams)
        self._n_inputs = len(self.upstreams)
        self._term_flag = Event()
        self._global_term_flag = None
        self._logger = None

    def __repr__(self):
        return self.name

    def __copy__(self):
        cls = self.__class__
        cp = cls.__new__(cls)
        cp.__dict__.update(self.__dict__)
        cp._term_flag = Event()
        return cp

    def _id(self):
        return str(id(self))

    def __hash__(self):
        return int(self._id())

    def __eq__(self, other):
        return self.__hash__() == hash(other)

    def set_logger(self, logger):
        self._logger = logger

    def set_term_flag(self, term_flag):
        self._term_flag = term_flag

    def set_global_term_flag(self, term_flag):
        self._global_term_flag = term_flag

    def set_upstreams(self, upstreams):
        self.upstreams = upstreams
        self._n_inputs = len(upstreams)

    def set_downstreams(self, downstreams):
        self.downstreams = downstreams
        self._n_outputs = len(downstreams)

    def get_upstreams(self):
        return self.upstreams

    def get_downstreams(self):
        return self.downstreams

    def public_vars(self):
        return {k:v for k,v in vars(self).items() if k[0] != '_'}

    def reset(self):
        self._n_rcvd_term_sigs = 0
        self._term_flag.clear()
        for stream in self.downstreams:
            stream.flush()

    def _in(self):
        x = [stream.get() for stream in self.upstreams]
        self._logger.log("in({})".format([repr(x_i) for x_i in x]), self.name, 'debug')
        return x

    def _out(self, x):
        if self._n_outputs <=1:
            x = [x]
        self._logger.log("out({})".format([repr(x_i) for x_i in x]), self.name, 'debug')
        for x_i, stream in zip(x, self.downstreams):
            stream.put(x_i)

    def _continue(self):
        return not (self._term_flag.is_set() or self._global_term_flag.is_set())

    def _contains_sentinel(self, x):
        for x_i in x:
            if x_i is Sentinel:
                return True
        return False

    def _contains_none(self, x):
        for x_i in x:
            if x_i is None:
                return True
        return False

    def _terminate_global(self):
        self._global_term_flag.set()
        self._logger.log("Global termination", self.name, "error")
        return True

    def _terminate_local(self):

        # Each time a Sentinel is caught this method is called
        # Pipe segment is only shutdown if its recieved a Sentinel from each Pipe upstream
        # If all upstream Pipes are accounted for:
        #    1) set term flag
        #    2) send 1 Sentinel to each downstream Pipe
        #    3) return True

        self._n_rcvd_term_sigs += 1
        if self._n_rcvd_term_sigs < self._n_ances:
            return False
        else:
            self._term_flag.set()
            for _ in range(self._n_desc):
                self._out([Sentinel]*self._n_outputs if self._n_outputs > 1 else Sentinel)
            self._logger.log("Local termination", self.name)
            return True

    def run_functor(self):
        # To be implemented in derived classes
        pass

    def run_pipe(self, name=None):
        # This method is called once on local process

        if name is not None:
            self.name = name

        # Check if local_init exists and if so initialize on local process
        if hasattr(self.functor_obj, 'local_init'):
            self.functor_obj.local_init(**self.init_kwargs)

        # Check if run method exists, if so use it as functor, otherwise use __call__ method
        if hasattr(self.functor_obj, 'run'):
            self.functor = self.functor_obj.run
        else:
            self.functor = self.functor_obj

        try:
            # run functor loop on local process
            self.run_functor()

            # check if local_term method exists, if so run once after pipe segment has terminated
            if hasattr(self.functor_obj, 'local_term'):
                self.functor_obj.local_term()

        except KeyboardInterrupt:
            self._logger.log("KeyboardInterrupt", self.name, 'error')





""" Derived Pipe classes """

from minipipe.base import Pipe, Sentinel

class Source(Pipe):
    """
        Source pipes are used to load and/or generate data. Sources have no upstreams, but will have one or more
        downstreams. Functor must be a valid Python generator. The generator should be initialized before passing
        as argument.

        :param functor: Python generator
        :param name: String associated with pipe segment
        :param downstreams: List of Streams that are outputs of functor
        :param ignore_exceptions: List of exceptions to ignore while pipeline is running
    """

    def __init__(self, functor_obj, name='Source', downstreams=None, ignore_exceptions=None):

        # Source has no upstreams
        super(Source, self).__init__(functor_obj, name, None, downstreams, ignore_exceptions)

    def run_functor(self):

        # Generator functors are used for sources
        # Will terminate once end of generator is reached

        x = None
        while self._continue():
            try:

                # Get next item from generator
                try:
                    x = next(self.functor)
                except TypeError:
                    # If generator is not initialized
                    self.functor = self.functor()
                    x = next(self.functor)

                # Check for Sentinel signaling termination
                if x is Sentinel:
                    self._terminate_local()
                    break

                # Do nothing on python None
                if x is None:
                    continue

                self._out(x)

            # Terminate once end of generator is reached
            except StopIteration:
                self._logger.log('End of stream', self.name)
                self._terminate_local()

            # These exceptions are ignored raising WARNING only
            except BaseException as e:
                if e.__class__ in self.ignore_exceptions:
                    self._logger.log(str(e), self.name, 'warning')
                    continue
                else:
                    self._logger.log(str(e), self.name, 'error')
                    self._terminate_global()
                    raise e


class Sink(Pipe):
    """
            Sink pipes are typically used to save/output data. Sinks have no downstreams, but will have one or more
            upstreams. Functor should be either a Python function or class with a "run" method and optionally
            "local_init" and "local_term" methods. Local_init, if supplied will be called once on the local process
            before run, while local_term will be called once afterwards.

            :param functor: Python function or class
            :param name: String associated with pipe segment
            :param upstreams: List of Streams that are inputs to functor
            :param init_kwargs: Kwargs to initiate class object on process (no used when func_type = 'function')
            :param ignore_exceptions: List of exceptions to ignore while pipeline is running

        """

    def __init__(self, functor_obj, name='Sink', upstreams=None,
                 ignore_exceptions=None, init_kwargs=None):

        # Sink has no downstreams
        super(Sink, self).__init__(functor_obj, name, upstreams, None,
                                   ignore_exceptions, init_kwargs)

    def run_functor(self):

        x = None
        while self._continue():
            try:

                x = self._in()

                # Check for Sentinel signaling termination
                if self._contains_sentinel(x):
                    if self._terminate_local():
                        break
                    else:
                        continue

                # Do nothing on python None
                if self._contains_none(x):
                    continue

                x = self.functor(*x)
                self._out(x)

            # These exceptions are ignored raising WARNING only
            except BaseException as e:
                if e.__class__ in self.ignore_exceptions:
                    self._logger.log(str(e), self.name, 'warning')
                    continue
                else:
                    self._logger.log(str(e), self.name, 'error')
                    self._terminate_global()
                    raise e


class Transform(Pipe):
    """
            Transform pipes are used to perform arbitrary transformations on data. Transforms will have one or more
            upstreams and downstreams. Functor should be either a Python function or class with a "run" method and optionally
            "local_init" and "local_term" methods. Local_init, if supplied will be called once on the local process
            before run, while local_term will be called once afterwards.

            :param functor: Python function or class
            :param name: String associated with pipe segment
            :param upstreams: List of Streams that are inputs to functor
            :param downstreams: List of Streams that are outputs of functor
            :param init_kwargs: Kwargs to initiate class object on process (no used when func_type = 'function')
            :param ignore_exceptions: List of exceptions to ignore while pipeline is running
        """

    def __init__(self, functor_obj, name='Transform', upstreams=None, downstreams=None,
                 ignore_exceptions=None, init_kwargs=None):

        super(Transform, self).__init__(functor_obj, name, upstreams, downstreams,
                                        ignore_exceptions, init_kwargs)

    def run_functor(self):

        x = None
        while self._continue():
            try:

                x = self._in()

                # Check for Sentinel signaling termination
                if self._contains_sentinel(x):
                    if self._terminate_local():
                        break
                    else:
                        continue

                # Do nothing on python None
                if self._contains_none(x):
                    continue

                x = self.functor(*x)
                self._out(x)

            # These exceptions are ignored raising WARNING only
            except BaseException as e:
                if e.__class__ in self.ignore_exceptions:
                    self._logger.log(str(e), self.name, 'warning')
                    continue
                else:
                    self._logger.log(str(e), self.name, 'error')
                    self._terminate_global()
                    raise e

class Regulator(Pipe):
    """
            Regulator pipes are a special type of transformation that changes the data chunk throughput, typically used
            for batching or accumulating data. Regulators can have both upstreams and downstreams. Functor should be a
            Python coroutine. The coroutine should not be initialized, instead use init_kwargs to initialize on the local
            process.

            :param functor: Python coroutines
            :param name: String associated with pipe segment
            :param upstreams: List of Streams that are inputs to functor
            :param downstreams: List of Streams that are outputs of functor
            :param init_kwargs: Kwargs to initiate class object on process (no used when func_type = 'function')
            :param ignore_exceptions: List of exceptions to ignore while pipeline is running
        """

    def __init__(self, functor_obj, name='Regulator', upstreams=None, downstreams=None,
                 ignore_exceptions=None, init_kwargs=None):

        super(Regulator, self).__init__(functor_obj, name, upstreams, downstreams,
                                        ignore_exceptions, init_kwargs)

    def run_functor(self):

        # Coroutine functors act as a transformation and source
        # Useful when the data needs to be broken up or accumulated
        # On StopIteration coroutine is reset

        coroutine = self.functor(**self.init_kwargs)
        next(coroutine)
        x = None
        while self._continue():
            try:

                x = self._in()

                # Check for Sentinel signaling termination
                if self._contains_sentinel(x):
                    if self._terminate_local():
                        break
                    else:
                        continue

                # Do nothing on python None
                if self._contains_none(x):
                    continue

                # Send data to coroutine
                x_i = coroutine.send(*x)

                # Iterate over coroutine output
                while x_i is not None:
                    self._out(x_i)
                    try:
                        x_i = next(coroutine)
                    except StopIteration:
                        # Reset coroutine for next data
                        coroutine = self.functor(**self.init_kwargs)
                        next(coroutine)
                        break

            # These exceptions are ignored raising WARNING only
            except BaseException as e:
                if e.__class__ in self.ignore_exceptions:
                    self._logger.log(str(e), self.name, 'warning')
                    continue
                else:
                    self._logger.log(str(e), self.name, 'error')
                    self._terminate_global()
                    raise e



from setuptools import setup

long_description = """
MiniPipe is a mini-batch pipeline designed for training machine learning models on very large datasets in a streaming 
fashion, written in pure python. MiniPipe is designed for situations where the data are too large to fit into memory, 
or when doing so would discourage experiment iterations due to prohibitively long loading and/or processing times.
"""

setup(name='minipipe',
      version='0.1.8',
      description='A machine learning mini-batch pipeline for out-of-memory training',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://minipipe.readthedocs.io',
      author='James D. Pearce',
      author_email='jdp.pearce@gmail.com',
      license='MIT',
      package_dir  = {'minipipe' : 'src'},
      packages=['minipipe'],
      install_requires=[
          'graphviz',
      ],
      classifiers = [
	    "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
	],
      zip_safe=False)



"""Minipipe has two APIs PipeLine and PipeSystem. PipeLine is for sequential pipelines while
PipeSystem can be used in any topology."""

from minipipe.pipes import Source
from minipipe.base import Logger, Stream
from multiprocessing import Process, Event
from collections import Counter
from graphviz import Digraph
from copy import copy

class PipeSystem(object):

    """
        PipeSystem connects Pipes and creates process pool. Pipes are run and closed with a built PipeSystem.

        Toy example:

        .. code-block:: python

            # Define functors
            def genRand(n=10):
                for _ in range(n):
                    yield np.random.rand(10)

            def batch(batch_size=2):
                x = (yield)
                for i in range(len(x)//batch_size):
                    yield x[i*batch_size:(i+1)*batch_size]

            def sumBatch(x):
                return x.sum()

            def split(x):
                return [x, None] if x > 1 else [None, x]

            def output_gt_1(x):
                print '1 <',x

            def output_lt_1(x):
                print '1 >',x

            # Define streams
            s1, s2, s3, s4, s5 = Stream(), Stream(), Stream(), Stream(), Stream()

            # Create Pipe segments with up/downstreams
            # Order is not important
            pipes = [
                Pipe(genRand, 'source1', downstreams=[s1], func_type='generator'),
                Pipe(genRand, 'source2', downstreams=[s1], func_type='generator'),
                Pipe(batch, 'batcher', upstreams=[s1], downstreams=[s2], func_type='coroutine'),
                Pipe(sumBatch, 'sum', upstreams=[s2], downstreams=[s3]),
                Pipe(sumBatch, 'sum', upstreams=[s2], downstreams=[s3]),
                Pipe(sumBatch, 'sum', upstreams=[s2], downstreams=[s3]),
                Pipe(split, 'split', upstreams=[s2], downstreams=[s4, s5]),
                Pipe(output_gt_1, 'print_gt_1', upstreams=[s4]),
                Pipe(output_lt_1, 'print_lt_1', upstreams=[s5]),
            ]

            # Build pipesystem
            psys = PipeSystem(pipes)
            psys.build()

            # Run pipesystem
            psys.run()
            psys.close()

    """

    def __init__(self, pipes):
        self.pipes = pipes
        self.streams = None
        self.processes = None
        self.built = False

    def build(self, log_lvl='INFO', monitor=False, ignore_exceptions=None):
        """Connects pipe segments together and builds graph."""

        self.log_lvl = log_lvl
        self.monitor = monitor
        self.ignore_exceptions = ignore_exceptions
        self.logger = Logger(log_lvl)
        self.global_term = Event()

        # Handle name collisions
        pnames = Counter([p.name for p in self.pipes])
        for pname, cnt in pnames.items():
            if cnt > 1:
                p_with_collisions = filter(lambda x: x.name==pname, self.pipes)
                for i, p in enumerate(p_with_collisions):
                    p.name += '_{}'.format(i)

        # Connect graph and set global term flag
        for p in self.pipes:
            if self.ignore_exceptions is not None:
                p.ignore_exceptions = self.ignore_exceptions
            p.set_global_term_flag(self.global_term)
            #self._connect_upstreams(p)
            #self._connect_downstreams(p)
            self._connect_streams(p)

        # For each pipe count all upstreams and downstreams.
        # Counts are used to determine the number of sentinels a pipe should receive before terminating.
        for p in self.pipes:
            self._count_relatives(p)

        # Create process pool
        self.processes = [Process(target=p.run_pipe, args=[p.name], name=p.name)
                          for p in self.pipes]

        # Add logger to each pipe
        for p in self.pipes:
            p.set_logger(self.logger)
            for s in p.downstreams + p.upstreams:
                if (s.monitor or self.monitor) and s.logger is None:
                    s.set_logger(self.logger)

        self.built = True

    def run(self):
        """Runs pipeline."""
        if self.processes is None:
            self.build()
        for proc in self.processes:
            proc.start()

    def close(self):
        """Joins pipeline."""
        for proc in self.processes:
            proc.join()

    def reset(self):
        """Resets pipeline."""
        self.global_term.clear()
        for p in self.pipes:
            p.reset()
            p.set_logger(self.logger)

        # Create new processes since they can only be started once
        self.processes = [Process(target=p.run_pipe, args=[p.name], name=p.name)
                          for p in self.pipes]

    #def _connect_downstreams(self, p):
    #    for stream in p.get_downstreams():
    #        stream.add_pipe_out(p)
            #for p_in in stream.pipes_in:
            #    p_in._n_ances += 1

    #def _connect_upstreams(self, p):
    #    for stream in p.get_upstreams():
    #        stream.add_pipe_in(p)
            #for p_out in stream.pipes_out:
            #    p_out._n_desc += 1

    def _connect_streams(self, p):
        for stream in p.get_downstreams():
            stream.add_pipe_out(p)
        for stream in p.get_upstreams():
            stream.add_pipe_in(p)

    def _count_relatives(self, p):
        for stream in p.get_downstreams():
            for p_in in stream.pipes_in:
                p_in._n_ances += 1
        for stream in p.get_upstreams():
            for p_out in stream.pipes_out:
                p_out._n_desc += 1


    def diagram(self, draw_streams=False):
        """Draws a graph diagram of pipeline."""

        assert(self.built==True), "ERROR: PipeSystem must be built first"
        g = Digraph()
        edges = set()

        # Assumes graph is a DAG thus iterate over downstreams only
        # There can only be one edge between nodes
        for p in self.pipes:
            g.node(p._id(), p.name)
            for s in p.get_downstreams():
                if draw_streams:
                    g.node(s._id(), s.name, shape='rectangle')
                    edge = (p._id(), s._id())
                    if edge not in edges:
                        edges.add(edge)
                        g.edge(*edge)
                for p_in in s.pipes_in:
                    g.node(p_in._id(), p_in.name)
                    if draw_streams:
                        edge = (s._id(), p_in._id())
                    else:
                        edge = (p._id(), p_in._id())
                    if edge not in edges:
                        edges.add(edge)
                        g.edge(*edge)
        return g


class PipeLine(PipeSystem):

    """
        A simplified API for linear PipeSytems.

        Toy example:

        .. code-block:: python

            # Define functors
            def genRand(n=10):
                for _ in range(n):
                    yield np.random.rand(10)

            def batch(batch_size=2):
                x = (yield)
                for i in range(len(x)//batch_size):
                    yield x[i*batch_size:(i+1)*batch_size]

            def sumBatch(x):
                return x.sum()

            def print_out(x):
                print (x)

            # Define pipeline
            pline = PipeLine()
            pline.add(Pipe(genRand, 'source1', func_type='generator'))
            pline.add(Pipe(batch, 'batcher', func_type='coroutine'), buffer_size = 10)
            pline.add(Pipe(sumBatch, 'sum'), n_processes = 3)
            pline.add(Pipe(print_out, 'print'))

            # Build pipeline
            pline.build()

            # Run pipeline
            pline.run()
            pline.close()

    """

    def __init__(self, monitor=False):
        self.monitor = monitor
        self.segments = []
        self.pipes = []
        self.sid = 0
        super(PipeLine, self).__init__(self.pipes)

    def add(self, pipe, n_processes=1, buffer_size=3):
        """Adds a pipe segment to the pipeline.

           :param pipe: Pipe segment to add to PipeLine
           :param n_processes: Number of processes (workers) to assign to pipe segment
           :param buffer_size: Size of Stream buffer
           :return: None
        """
        # Python generators cannot be split
        assert not (isinstance(pipe, Source) and n_processes > 1), 'Use PipeSystem API for multiple Sources'

        # After the Source connect each segment with a Stream
        if len(self.segments) > 0:
            s = Stream(buffer_size=buffer_size,
                       name = 'stream_{}'.format(self.sid),
                       monitor=self.monitor)
            self.sid += 1
            for seg in self.segments[-1]:
                seg.set_downstreams([s])
            pipe.set_upstreams([s])

        # Create a copy of the pipe segment for each process
        seg = [copy(pipe) for _ in range(n_processes)]

        self.pipes += seg
        self.segments.append(seg)